package rendering

import (
	"context"
	"errors"
	"fmt"
	"sort"
)

// ErrDuplicateName is returned by EngineRegistry.Register if and engine with
// the same name is already registered
var ErrDuplicateName = errors.New("duplicate engine name")

// EngineTransitionInfo describes a
type EngineTransitionInfo struct {
	FromType string
	ToType   string
	Priority float64
}

// EngineMetadataInfo ...
type EngineMetadataInfo struct {
	Name        string
	Default     string
	Description string
}

// EngineInfo describes an engine capabilities
type EngineInfo struct {
	Type        string
	Description string
	Transitions []EngineTransitionInfo
	Metadata    []EngineMetadataInfo
}

// Engine is able to render a document
type Engine interface {
	Info() EngineInfo
	Render(ctx context.Context, w DocumentWriter, r DocumentReader, toType string) error
}

// RegisterEngine a new engine factory
func RegisterEngine(name string, engine Engine) error {
	return DefaultEngineRegistry.Register(name, engine)
}

// MustRegisterEngine registers an Engine and panics if it fails
func MustRegisterEngine(name string, engine Engine) {
	DefaultEngineRegistry.MustRegister(name, engine)
}

// EngineRegistry is a set of Engine, ready for use
type EngineRegistry struct {
	Engines map[string]Engine
}

// Register a Engine
func (s *EngineRegistry) Register(name string, engine Engine) error {
	if _, ok := s.Engines[name]; ok {
		return ErrDuplicateName
	}
	s.Engines[name] = engine
	return nil
}

// MustRegister a Engine
func (s *EngineRegistry) MustRegister(name string, engine Engine) {
	if err := s.Register(name, engine); err != nil {
		panic(err)
	}
}

// PipelineStep is a step in a pipeline
type PipelineStep struct {
	TargetType string
	Engine     Engine
}

type pipelineStepResult struct {
	step PipelineStep
	err  error
}

type Pipeline []PipelineStep

func (p Pipeline) Render(ctx context.Context, w DocumentWriter, r DocumentReader) error {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	if len(p) == 0 {
		return fmt.Errorf("cannot render a empty pipeline")
	}

	results := make(chan pipelineStepResult, len(p))

	runStep := func(step PipelineStep, w DocumentWriter, r DocumentReader) {
		w.SetType(step.TargetType)
		result := pipelineStepResult{
			step: step,
			err:  step.Engine.Render(ctx, w, r, step.TargetType),
		}
		if err := w.Close(); err != nil {
			if result.err == nil {
				result.err = err
			}
		}
		results <- result
	}

	for _, step := range p[:len(p)-1] {
		pw, pr := DocumentPipe()
		go runStep(step, pw, r)
		r = pr
	}

	go runStep(p[len(p)-1], w, r)

	var err error

	for i := 0; i != len(p); i++ {
		stepResult := <-results
		if stepResult.err != nil {
			if err == nil {
				err = fmt.Errorf(
					"engine %s failed: %s",
					stepResult.step.Engine.Info().Type, stepResult.err)
				cancel()
			}
		}
	}

	return err
}

// GetPipeline returns a list of engine for converting a document of type 'fromType'
// to type 'toType'
// TODO do a real implementation, not this hard-coded hack
func (s *EngineRegistry) GetPipeline(fromType, toType string) (Pipeline, error) {
	if fromType == toType {
		return nil, nil
	}
	if fromType == "text/mjml" && toType == "application/pdf" {
		engine1, err := s.FindEngine(fromType, "text/html")
		if err != nil {
			return nil, err
		}
		engine2, err := s.FindEngine("text/html", toType)
		if err != nil {
			return nil, err
		}
		return []PipelineStep{
			{"text/html", engine1},
			{"application/pdf", engine2},
		}, nil
	}
	engine, err := s.FindEngine(fromType, toType)
	if err != nil {
		return nil, err
	}
	return Pipeline{
		{toType, engine},
	}, nil
}

// FindEngine finds the engine to do a transformation that has the highest priority
func (s *EngineRegistry) FindEngine(fromType, toType string) (Engine, error) {
	engines, err := s.FindEngines(fromType, toType)
	if err != nil {
		return nil, err
	}
	if len(engines) == 0 {
		return nil, fmt.Errorf("no engine found to convert from %s to %s", fromType, toType)
	}
	return engines[0], nil
}

// FindEngines finds engines that can do a given transformation. The result
// is sorted by priority, the highest first
func (s *EngineRegistry) FindEngines(fromType, toType string) ([]Engine, error) {
	type entry struct {
		engine   Engine
		priority float64
	}

	entries := []entry{}
	for _, engine := range s.Engines {
		for _, tr := range engine.Info().Transitions {
			if tr.FromType == fromType && tr.ToType == toType {
				entries = append(entries, entry{engine, tr.Priority})
			}
		}
	}
	sort.Slice(entries, func(i, j int) bool {
		return entries[i].priority > entries[j].priority
	})
	var engines = make([]Engine, 0, len(entries))
	for _, entry := range entries {
		engines = append(engines, entry.engine)
	}
	return engines, nil
}

// NewEngineRegistry creates a EngineRegistry
func NewEngineRegistry() *EngineRegistry {
	return &EngineRegistry{
		Engines: make(map[string]Engine),
	}
}

var (
	// DefaultEngineRegistry is the default engine registry
	DefaultEngineRegistry = NewEngineRegistry()
)
